#!/usr/bin/env python3
'''
MLflow utility functions.
'''

import logging
import os
from contextlib import ExitStack
from types import TracebackType
from typing import Any, Optional, Type

import mlflow
from hydra.core.hydra_config import HydraConfig
from hydra.utils import to_absolute_path
from omegaconf import DictConfig, ListConfig

from hydronaut.hydra.omegaconf import get, get_container


LOGGER = logging.getLogger(__name__)


def _recursively_log_omegaconf_params_with_mlflow(name: str, obj: Any) -> None:
    '''
    Recursively log the Hydra parameters.

    Args:
        name:
            The parameter name.

        obj:
            The parameter object.
    '''
    if isinstance(obj, DictConfig):
        for key, value in obj.items():
            _recursively_log_omegaconf_params_with_mlflow(f'{name}.{key}', value)
    elif isinstance(obj, ListConfig):
        for i, value in enumerate(obj):
            _recursively_log_omegaconf_params_with_mlflow(f'{name}.{i:d}', value)
    else:
        LOGGER.debug('logging parameter: %s = %s', name, obj)
        mlflow.log_param(name, obj)


class MLflowRunner():
    '''
    Context manager to set up an MLflow run from an Omegaconf configuration
    object. When used as a context manager, it will configure the tracking URI,
    experiment, artifact location and run parameters from the configurat object.
    The context will return the MLflow active run.
    '''
    def __init__(self, config: DictConfig) -> None:
        '''
        Args:
            config:
                The Omegaconf configuration object.
        '''
        self.config = config
        self.exit_stack = ExitStack()

    def log_parameters(self) -> None:
        '''
        Log all parameters with MLflow.
        '''
        for name, value in self.config.items():
            _recursively_log_omegaconf_params_with_mlflow(name, value)

    def _set_experiment(self) -> mlflow.entities.Experiment:
        '''
        Set the current experiment from the name in the configuration object.

        Returns:
            An instance of mlflow.entities.Experiment.
        '''
        config = self.config

        exp_name = config.experiment.name
        LOGGER.info('MLflow experiment name: %s', exp_name)

        exp = mlflow.get_experiment_by_name(exp_name)
        if exp is None:
            exp_id = mlflow.create_experiment(name=exp_name)
            exp = mlflow.get_experiment(exp_id)
        mlflow.set_experiment(experiment_id=exp.experiment_id)

        return exp

    def _set_uris(self) -> None:
        '''
        Log and set the MLflow URIs.
        '''
        tracking_uri = mlflow.get_tracking_uri()
        LOGGER.info('MLflow tracking URI: %s', tracking_uri)

        registry_uri = os.getenv('MLFLOW_REGISTRY_URI', default=tracking_uri)
        if registry_uri:
            mlflow.set_registry_uri(registry_uri)

        LOGGER.info(
            'MLflow registry URI: %s',
            mlflow.get_registry_uri()
        )

    def __enter__(self) -> mlflow.ActiveRun:
        config = self.config
        hydra_config = HydraConfig.get()

        stack = self.exit_stack
        self._set_uris()
        self._set_experiment()

        run_kwargs = get_container(self.config, 'experiment.mlflow.start_run', default={})

        run_name = config.experiment.name
        job_id = hydra_config.job.get('id')
        if job_id:
            run_name = f'{run_name}-{job_id}'
        run_kwargs['run_name'] = run_name

        for key, value in sorted(run_kwargs.items()):
            LOGGER.debug('MLflow start_run parameter: %s = %s', key, value)

        active_run = stack.enter_context(mlflow.start_run(**run_kwargs))

        LOGGER.info(
            'MLflow artifact URI: %s',
            mlflow.get_artifact_uri()
        )

        self.log_parameters()
        return active_run

    def __exit__(
        self,
        typ: Optional[Type[BaseException]],
        value: Optional[BaseException],
        traceback: Optional[TracebackType]
    ) -> bool:
        self.exit_stack.close()



#!/usr/bin/env python3

'''
File functions.
'''

import contextlib
import io
import logging
import pathlib
import shutil
import time
from typing import Any

import mlflow
from omegaconf import OmegaConf


LOGGER = logging.getLogger(__name__)


def backup_copy(path: pathlib.Path) -> None:
    '''
    Attempt to backup a back to a timestamped path in the same directory.

    Args:
        path:
            The path to back up. If it does not exist then nothing will be done.
    '''
    if path.exists():
        while True:
            timestamp = time.strftime('%Y_%m_%d_%H_%M_%S')
            new_suffix = f'.{timestamp}{path.suffix}'
            backup_path = path.with_suffix(new_suffix)
            if not backup_path.exists():
                break
        LOGGER.info('backing up %s to %s', path, backup_path)
        if path.is_dir():
            shutil.copytree(path, backup_path)
        else:
            shutil.copy2(path, backup_path)


@contextlib.contextmanager
def safe_open(path: pathlib.Path, *args: Any, backup: bool = False, **kwargs: Any) -> io.TextIOBase:
    '''
    Context manager for safely opening a path by first creating any missing
    parent directories and then moving any existing file to a backup path.

    Args:
        path:
            The path to open.

        backup:
            If True, backup the file at the path if it exists.

        *args, **kwargs:
            Positional and keyword arguments passed through to
            pathlib.Path.open.
    Returns:
        The open file handle.
    '''
    path = pathlib.Path(path).resolve()
    LOGGER.debug('opening %s', path)

    path.parent.mkdir(parents=True, exist_ok=True)

    if backup:
        backup_copy(path)

    kwargs.setdefault('encoding', 'utf-8')
    with path.open(*args, **kwargs) as handle:  # pylint: disable=unspecified-encoding
        yield handle


def save_config(
    config: OmegaConf,
    path: pathlib.Path,
    resolve: bool = False,
    overwrite: bool = False
) -> None:
    '''
    Save a config to a path.

    Args:
        config:
            An OmegaConf configuration object.

        path:
            The output path.

        resolve:
            Resolve all resolvers before saving the config.
    '''
    if path.exists() and not overwrite:
        LOGGER.warning('refusing to overwrite existing file: %s', path)
    LOGGER.debug('saving config to %s', path)
    path.parent.mkdir(parents=True, exist_ok=True)
    with path.open('w', encoding='utf-8') as handle:
        OmegaConf.save(config=config, f=handle, resolve=resolve)
    mlflow.log_artifact(path, artifact_path='hydronaut')



#!/usr/bin/env python3
'''
Type definitions and functions.
'''

import pathlib
from typing import Tuple, Union

# Generic number type.
Number = Union[int, float]

# Optimization value for the Optuna sweeper returned by hydronaut.run.Runner.__call__.
OptimizationValue = Union[Number, Tuple[Number, ...]]

# Generic path type for functions that can handle both pathlib Paths or strings.
Path = Union[str, pathlib.Path]



#!/usr/bin/env python3
'''
Hydra resolvers.
'''

import logging
import multiprocessing

try:
    import torch
except ImportError:
    torch = None
from omegaconf import OmegaConf
from hydra.core.utils import setup_globals as hydra_setup_globals


LOGGER = logging.getLogger(__name__)


def _get_divisor(args):
    '''
    Internal function to handle optional divisor to some resolvers.
    '''
    if len(args) > 0:
        divisor = int(args[0])
        if divisor > 0:
            return divisor
    return 1


def get_n_cpu(*args):
    '''
    Get the number of available CPUs. Optionally accept an integer argument
    which will be used to divide the result.

    Args:
        *args:
            An optional single integer argument. If given, the number of CPUs
            will be divided by this number.

    Returns:
        The number of available CPUs. If a divisior is given then this will
        always return a value of at least 1.
    '''
    n_cpu = multiprocessing.cpu_count()
    return max(n_cpu // _get_divisor(args), 1)


def get_n_gpu_pytorch(*args):
    '''
    Get the number of available PyTorch GPUs. Optionally accept an integer
    argument which will be used to divide the result.

    Args:
        *args:
            An optional single integer argument. If given, the number of CPUs
            will be divided by this number.

    Returns:
        The number of available PyTorch GPUs. If the number is greater than 0
        than a value of at least 1 will always be returned when an optional
        divisor is given.
    '''
    if torch is None:
        LOGGER.warning('PyTorch must be installed to get the number of PyTorch GPUs')
        return 0
    n_gpu = torch.cuda.device_count()
    if n_gpu < 1:
        return 0
    return max(n_gpu // _get_divisor(args), 1)


def register() -> dict[str, str]:
    '''
    Register custom resolvers.

    Returns:
        A dict mapping resolver names to their descriptions.
    '''

    resolvers = {}
    for name, desc, func in (
            (
                'n_cpu',
                'The number of available CPUs on the current system.',
                get_n_cpu
            ),
            (
                'n_gpu_pytorch',
                'The number of GPUs recognized by PyTorch (requires PyTorch).',
                get_n_gpu_pytorch
            ),
            (
                'max',
                'The maximum argument.',
                max
            ),
            (
                'min',
                'The minimum argument.',
                min
            )
    ):
        OmegaConf.register_new_resolver(name, func)
        resolvers[name] = desc
    return resolvers


def reregister() -> None:
    '''
    Reregister custom and Hydra resolvers. This is necessary due to a bug in the
    process launcher.
    '''
    if not OmegaConf.has_resolver('n_cpu'):
        register()
    if not OmegaConf.has_resolver('hydra'):
        hydra_setup_globals()



#!/usr/bin/env python3
'''
Omegaconf utility functions.
'''

from typing import Any

from omegaconf import OmegaConf, DictConfig  # pylint: disable=import-self


def get(config: DictConfig, name: str, default: Any = None) -> Any:
    '''
    Get a nested parameter such as foo.bar.baz from an Omegaconf configuration
    object, with default value.

    Args:
        config:
            The configuration object.

        name:
            The nested parameter name. It will be split on ".".

        default:
            An optional default value. Default: None

    Returns:
        The parameter value, or the default value if any of the nested values
        are missing or the resulting value is None.
    '''
    value = config
    for component in name.split('.'):
        try:
            value = getattr(value, component)
        except AttributeError:
            return default
    if value is None:
        return default
    return value


def get_container(
    config: DictConfig,
    name: str,
    *,
    default: Any = None,
    resolve: bool = True
) -> Any:
    '''
    A wrapper around get() that transforms the result with OmegaConf.to_container.

    Args:
        config:
            Same as get().

        name:
            Same as get().

        default:
            Same as get().

        resolve:
            Passed through to to_container(). If True, the resolvers will be
            resolved before creating the container.

    Returns:
        Same as get().
    '''
    result = get(config, name, default=default)
    if OmegaConf.is_config(result):
        result = OmegaConf.to_container(result, resolve=resolve)
    return result



#!/usr/bin/env python3
'''
Sphinx configuration file.
'''

# pylint: disable=invalid-name,redefined-builtin

import importlib
import pathlib
import sys

source_code = '../../src'
git_url = 'https://gitlab.inria.fr/jrye/hydronaut'

this_path = pathlib.Path(__file__).resolve()
sys.path.insert(0, str((this_path.parent / source_code).resolve()))

author = 'Jan-Michael Rye'
copyright = '2022, Inria'
project = 'Hydronaut'
html_theme = 'sphinx_rtd_theme'

autodoc_mock_imports = [
    'hydra',
    'mlflow',
    'omegaconf',
    'pytorch_lightning',
    'torch'
]
extensions = [
  'sphinx.ext.autodoc',
  'sphinx.ext.napoleon',
  'sphinx.ext.linkcode',
  'sphinx.ext.todo',
  'myst_parser'
]


def linkcode_resolve(domain, info):
    '''Get source links for the linkcode extension.'''
    module = info['module']
    if domain != 'py' or not module:
        return None
    top_mod = importlib.import_module(module.split('.')[0])
    mod = importlib.import_module(module)
    top_mod_path = pathlib.Path(top_mod.__file__)
    mod_path = pathlib.Path(mod.__file__)
    subpath = str(mod_path.relative_to(top_mod_path.parent.parent))
    return f'{git_url}/-/blob/main/src/{subpath}'



#!/usr/bin/env python3
'''
Run an experiment with Hydra, Optuna and MLflow.

TODO
* Explore mlflow.autolog support in PyTorch.
'''

import importlib
import inspect
import logging
import os
import subprocess
import sys
from typing import Any

import hydra
import mlflow

from hydronaut.experiment import Experiment
from hydronaut.hydra.config import configure_hydra, configure_environment, HYDRA_VERSION_BASE
from hydronaut.hydra.omegaconf import get, get_container
from hydronaut.mlflow import MLflowRunner
from hydronaut.paths import PathManager
from hydronaut.types import Path, OptimizationValue

LOGGER = logging.getLogger(__name__)


class Runner():
    '''
    Experiment runner.
    '''

    def __init__(self, config_subpath: Path = None) -> None:
        '''
        Args:
            config_subpath:
                The subpath to the configuration file. This is
                passed through to hydronaut.paths.get_config_path
                to get the main Hydra configuration file.
        '''
        self._path_manager = PathManager()
        self.config_subpath = config_subpath
        # The Hydra configuration object, which will be set when this is run.
        self.config = None

    def get_experiment_object(self) -> Experiment:
        '''
        Get the experiment class specified by the configuration object.

        Returns:
            The experiment instance.

        Raises:
            ImportError:
                The module could not be found.

            AttributeError:
                The module does not contain the expected class name.
        '''
        config = self.config
        exp_cls_param = config.experiment.exp_class
        exp_cls_module, exp_cls_name = exp_cls_param.split(':', 1)
        importlib.import_module(exp_cls_module)
        exp_cls = getattr(sys.modules[exp_cls_module], exp_cls_name)
        LOGGER.debug('loaded experiment class: %s', exp_cls)

        if not issubclass(exp_cls, Experiment):
            LOGGER.warning(
                'Experiment class %s.%s [%s] is not a subclass of %s.%s',
                exp_cls.__module__,
                exp_cls.__qualname__,
                inspect.getfile(exp_cls),
                Experiment.__module__,
                Experiment.__qualname__
            )
        return exp_cls(config)

    def _configure_environment(self) -> None:
        '''
        Set environment variables defined in the configuration file.
        '''
        env_vars = get_container(self.config, 'experiment.environment', default={})
        for name, value in env_vars.items():
            os.environ[name] = value

    def __call__(self, *args: Any, **kwargs: Any) -> OptimizationValue:
        '''
        Run the experiment and return a value for the Optuna sweeper.

        Args:
            *args, **kwargs:
                Positional and keyword arguments passed through to the
                hydra.main function.

        Returns:
            The value to optimize.
        '''
        configure_hydra()
        config_path = self._path_manager.get_config_path(subpath=self.config_subpath)
        if not config_path.exists():
            LOGGER.error(
                # One string intentionally split across several lines.
                'The expected configuration file (%s) does exist. If you wish to '
                'use a different file, set the %s environment variable to the '
                'name of the file that you wish to use. Relative paths will be '
                'interpreted relative to %s.',
                config_path,
                self._path_manager.HYDRONAUT_CONFIG_ENV_VAR,
                self._path_manager.config_dir
            )
            return 1

        @hydra.main(
            version_base=HYDRA_VERSION_BASE,
            config_path=str(config_path.parent),
            config_name=config_path.stem
        )
        def _run(config):
            '''
            Internal runner function. This is defined within this method so that
            class attributes can be used as arguments to hydra.main().

            Args:
                config:
                    The Hydra configuration object. It must contain a field
                    named "experiment.class" that points to an importable Python
                    class which is a subclass of
                    hydronaut.experiment.Experiment.
            '''
            LOGGER.info('Hydronaut configuration file: %s', config_path)
            configure_hydra()
            self.config = config
            self._configure_environment()

            with MLflowRunner(config):
                # Configure the environment variables so that any experiment
                # subprocesses can invoke configure_hydra(from_env=True) to
                # re-establish the Hydra configuration. This also ensures that they
                # set the right MLflow run ID.
                configure_environment()

                self._path_manager.add_python_paths(
                    get(self.config, 'experiment.python.paths')
                )

                exp = self.get_experiment_object()
                exp.setup()
                obj_val = exp()
                mlflow.log_metric("Objective Value", obj_val)
                return obj_val
        return _run(*args, **kwargs)


def main(*args: Any, **kwargs: Any) -> OptimizationValue:
    '''
    Main function to run an experiment with the Hydra configuration.

    Args:
        *args, **kwargs:
            Positional and keyword arguments passed through to the hydra.main
            function.

    Returns:
        The value returned by calling the runner.
    '''
    return Runner()(*args, **kwargs)


def script_main() -> None:
    '''
    Hydra makes some assumptions about configuration paths based on how the main
    function is called. This is a workaround for creating a script via
    pyproject.toml. It will simply invoke main() with any passed command-line
    arguments.
    '''
    cmd = (sys.executable, '-m', 'hydronaut.run', *sys.argv[1:])
    try:
        subprocess.run(cmd, check=True)
    except subprocess.CalledProcessError as err:
        sys.exit(err)


if __name__ == '__main__':
    main()
